import os
import sys
from vipy.util import try_import, islist, tolist, tempdir, remkdir, chunklistbysize, listpkl, filetail, filebase, tempdir
from itertools import repeat
import numpy as np
import vipy.globals
from vipy.globals import log


try_import('dill','dill'); import dill
dill.extend(False)  # https://github.com/uqfoundation/dill/issues/383
try_import('dask', 'dask distributed')
from dask.distributed import as_completed, wait
import dask
import dask.config
from dask.distributed import Client
from dask.distributed import as_completed, wait
from dask.distributed import get_worker         
dill.extend(True)  # https://github.com/uqfoundation/dill/issues/383


class Batch():
    """vipy.batch.Batch class

    This class provides a representation of a set of vipy objects.  All of the object types must be the same.  If so, then an operation on the batch is performed on each of the elements in the batch in parallel.

    Examples:

    >>> b = vipy.batch.Batch([Image(filename='img_%06d.png' % k) for k in range(0,100)])
    >>> b.map(lambda im: im.bgr())  
    >>> b.map(lambda im: np.sum(im.array())) 
    >>> b.map(lambda im, f: im.saveas(f), args=['out%d.jpg' % k for k in range(0,100)])
    
    >>> v = vipy.video.RandomSceneActivity()
    >>> b = vipy.batch.Batch(v, n_processes=16)
    >>> b.map(lambda v,k: v[k], args=[(k,) for k in range(0, len(v))])  # paralle interpolation

    >>> d = vipy.data.kinetics.Kinetics700('/path/to/kinetics').download().trainset()
    >>> b = vipy.batch.Batch(d, n_processes=32)
    >>> b.map(lambda v: v.download().save())  # will download and clip dataset in parallel

    >>> b.result()  # retrieve results after a sequence of map or filter chains
    >>> list(b)     # equivalent to b.result()

    Args:
        strict: [bool] if distributed processing fails, return None for that element and print the exception rather than raise
        as_completed: [bool] Return the objects to the scheduler as they complete, this can introduce instabilities for large complex objects, use with caution
        ordered: [bool]: If True, then preserve the order of objects in objlist in distributed processing

    .. notes:: `vipy.dataset.Dataset.map` supports batch processing and is the preferred method for paralle processing of a dataset

    """    
             
    def __init__(self, objlist, strict=False, as_completed=False, warnme=False, minscatter=None, ordered=False):
        """Create a batch of homogeneous vipy.image objects from an iterable that can be operated on with a single parallel function call
        """
        assert isinstance(objlist, list), "Input must be a list"
        self._batchtype = type(objlist[0]) if len(objlist)>0 else type(None)
        assert all([isinstance(im, self._batchtype) for im in objlist]), "Invalid input - Must be homogeneous list of the same type"                

        # Move this into map and disable using localmap
        if vipy.globals.dask() is None and warnme:
            log.info('[vipy.batch.Batch]: vipy.batch.Batch() is not set to use parallelism.  This is set using:\n    >>> with vipy.globals.parallel(n) for multi-processing with n processes\n    >>> vipy.globals.parallel(pct=0.8) for multiprocessing that uses a percentage of the current system resources\n    >>> vipy.globals.dask(address="SCHEDULER:PORT") which connects to a Dask distributed scheduler.\n    >>> vipy.globals.noparallel() to completely disable all parallelism.')

        self._strict = strict
        self._as_completed = as_completed  # this may introduce instabilities for large complex objects, use with caution
        self._minscatter = minscatter
        self._ordered = ordered
        self._objlist = [(k,o) for (k,o) in enumerate(objlist)] if ordered else objlist

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        return self
    
    def __len__(self):
        return len(self._objlist)

    def __repr__(self):        
        return str('<vipy.batch: type=%s, len=%d%s>' % (str(self._batchtype), len(self), (', client=%s' % str(vipy.globals.dask())) if vipy.globals.dask() is not None else ''))

    def _client(self):
        return vipy.globals.dask()._client if vipy.globals.dask() is not None else None
    
    def _batch_wait(self, futures):
        try:
            results = []            
            for (k, batch) in enumerate(as_completed(futures, with_results=True, raise_errors=False).batches()):
                for (future, result) in batch:
                    if future.status != 'error':
                        results.append(result)  # not order preserving, will restore order in result()
                    else:
                        if self._strict:
                            typ, exc, tb = result
                            raise exc.with_traceback(tb)
                        else:
                            log.warning('[vipy.batch]: future %s failed with error "%s" - SKIPPING' % (str(future), str(result)))
                        results.append(None)
                    del future, result  # distributed memory cleanup

                # Distributed memory cleanup
                del batch
  
            return results

        except KeyboardInterrupt:
            # warnings.warn('[vipy.batch]: batch cannot be restarted after killing with ctrl-c - You must create a new Batch()')
            #vipy.globals.dask().shutdown()
            #self._client = None
            return results
        except:
            # warnings.warn('[vipy.batch]: batch cannot be restarted after exception - Recreate Batch()')                
            raise
    
    def _wait(self, futures):
        assert islist(futures) and all([hasattr(f, 'result') for f in futures])
        if self._as_completed:
            return self._batch_wait(futures)        
        try:
            results = []            
            wait(futures)
            for f in futures:  
                try:
                    results.append(f.result()) 
                except:
                    if self._strict:
                        raise
                    try:
                        log.error('[vipy.batch]: future %s failed with error "%s" for batch "%s"' % (str(f), str(f.exception()), str(self)))
                    except:
                        log.error('[vipy.batch]: future failed')
                    results.append(None)
            return results

        except KeyboardInterrupt:
            # warnings.warn('[vipy.batch]: batch cannot be restarted after killing with ctrl-c - You must create a new Batch()')
            #vipy.globals.dask().shutdown()
            #self._client = None
            return None  
        except:
            # warnings.warn('[vipy.batch]: batch cannot be restarted after exception - Recreate Batch()')                
            raise

    def result(self):
        """Return the result of the batch processing, ordered"""
        if self._ordered:
            objlist = {int(v[0]):v[1] for v in self._objlist if v is not None}
            return [objlist[k] if k in objlist else None for k in range(len(self._objlist))]  # restore order
        else:
            return self._objlist

    def __iter__(self):
        for x in self.result():
            yield x
            
    def map(self, f_lambda, args=None):
        """Run the lambda function on each of the elements of the batch and return the batch object.
        
        >>> iml = [vipy.image.RandomScene(512,512) for k in range(0,1000)]   
        >>> imb = vipy.image.Batch(iml) 
        >>> imb.map(lambda im: im.rgb())  

        The lambda function f_lambda should not include closures.  If it does, construct the lambda with default parameter capture:

        >>> f = lambda x, prm1=42: x+prm1

        instead of:

        >>> prm1 = 42
        >>> f = lambda x: x+prm1

        """
        c = self._client()

        if c is None:
            f_lambda_ordered = (lambda x,f=f_lambda: (x[0], f(x[1]))) if self._ordered else f_lambda
            self._objlist = [f_lambda_ordered(o) for o in self._objlist]  # no parallelism
        else:
            f_lambda_ordered = (lambda x,f=f_lambda: (x[0], f(x[1]))) if self._ordered else f_lambda
            objlist = c.scatter(self._objlist) if (self._minscatter is not None and len(self._objlist) >= self._minscatter) else self._objlist
            self._objlist = self._wait(c.map(f_lambda_ordered, objlist))
        return self

    def filter(self, f_lambda):
        """Run the lambda function on each of the elements of the batch and filter based on the provided lambda keeping those elements that return true 
        """
        c = self._client()
        f_lambda_ordered = (lambda x,f=f_lambda: (x[0], f(x[1]))) if self._ordered else f_lambda

        if c is None:
            self._objlist = [o for o in self._objlist if f_lambda_ordered(o)[1]]  # no parallelism
        else:
            objlist = self._objlist  # original list
            is_filtered = self._wait(c.map(f_lambda_ordered, c.scatter(self._objlist)))  # distributed filter (replaces self._objlist)
            self._objlist = [obj for (f, obj) in zip(is_filtered, objlist) if f[1] is True]  # keep only elements that filter true
        return self
        
    def scattermap(self, f_lambda, obj):
        """Scatter obj to all workers, and apply lambda function f(obj, im) to each element in batch
        
           Usage: 
         
           >>> Batch(mylist, ngpu=8).scattermap(lambda net, im: net(im), net).result()
        
           This will scatter the large object net to all workers, and pin it to a specific GPU.  Within the net object, you can call 
           vipy.global.gpuindex() to retrieve your assigned GPU index, which can be used by torch.cuda.device().  Then, the net
           object processes each element in the batch using net according to the lambda, and returns the results.  This function 
           includes ngpu processes, and assumes there are ngpu available on the target machine.  Each net is replicated in a different
           process, so it is the callers responsibility for getting vipy.global.gpuindex() from within the process and setting 
           net to take advantage of this GPU rather than using the default cuda:0.  

        """
        c = self._client()
        f_lambda_ordered = (lambda net,x,f=f_lambda: (x[0], f(net,x[1]))) if self._ordered else (lambda net,x,f=f_lambda: f(net, x))

        if c is None:
            self._objlist = [f_lambda_ordered(obj, o) for o in self._objlist]  # no parallelism
        else:
            objdist = c.scatter(obj, broadcast=True)        
            objlist = c.scatter(self._objlist) if (self._minscatter is not None and len(self._objlist) >= self._minscatter) else self._objlist
            self._objlist = self._wait([c.submit(f_lambda_ordered, objdist, im) for im in objlist])
        return self





import os
import vipy
from vipy.util import filetail


def open_images_v7(datadir=None):
    """https://storage.googleapis.com/openimages/web/download_v7.html
     https://github.com/cvdfoundation/open-images-dataset#download-full-dataset-with-google-storage-transfer
     https://github.com/Tencent/tencent-ml-images?tab=readme-ov-file#download-images

    ```python
     trainset = open_images_v7()
     d = trainset.take(16).map(lambda im: im.load(ignoreErrors=True)).filter(lambda im: im.isloaded())
     vipy.visualize.montage(d.map(lambda im: im.mindim(512).centersquare().annotate()).list()).show()
    ```

    This returns the tagged image and bounding boxes for the open_images v7 trainset
    
    This can take a long while to parse due to large csv files

    """

    urls = ['https://storage.googleapis.com/openimages/v6/oidv6-train-annotations-bbox.csv',
            'https://storage.googleapis.com/openimages/v5/train-annotations-human-imagelabels-boxable.csv',
            'https://storage.googleapis.com/openimages/v5/train-annotations-human-imagelabels.csv',    
            'https://storage.googleapis.com/openimages/2018_04/train/train-images-boxable-with-rotation.csv',
            'https://storage.googleapis.com/openimages/v7/oidv7-class-descriptions-boxable.csv',
            'https://storage.googleapis.com/openimages/v7/oidv7-classes-trainable.txt',
            'https://storage.googleapis.com/openimages/v7/oidv7-class-descriptions.csv',
            'https://storage.googleapis.com/openimages/v6/oidv6-class-descriptions.csv',
            'https://storage.googleapis.com/openimages/v6/oidv6-train-images-with-labels-with-rotation.csv']
    
    labeldir = vipy.util.remkdir(vipy.util.tocache('open_images_v7') if not datadir else datadir)

    for url in urls:
        if not os.path.exists(os.path.join(labeldir, filetail(url))):
            vipy.downloader.download(url, os.path.join(labeldir, filetail(url)))
        
    trainable = set(vipy.util.readlist(os.path.join(labeldir, 'oidv7-classes-trainable.txt')))
    d_label_to_category = {r[0]:r[1] for r in vipy.util.readcsv(os.path.join(labeldir, 'oidv6-class-descriptions.csv'), ignoreheader=True)}
    d_label_to_category |= {r[0]:r[1] for r in vipy.util.readcsv(os.path.join(labeldir, 'oidv7-class-descriptions.csv'), ignoreheader=True)}
    d_label_to_object_category = {r[0]:r[1] for r in vipy.util.readcsv(os.path.join(labeldir, 'oidv7-class-descriptions-boxable.csv'), ignoreheader=True)}

    d_train_imageid_to_labels = vipy.util.groupbyasdict([(r[0], r[2]) for r in vipy.util.readcsv(os.path.join(labeldir, 'train-annotations-human-imagelabels.csv'), ignoreheader=True) if float(r[3])>0], keyfunc=lambda x: x[0], valuefunc=lambda x: x[1])
    d_train_imageid_to_objects = vipy.util.groupbyasdict([(r[0], r[2], float(r[4]), float(r[6]), float(r[5]), float(r[7]))  # (imageid, category, xmin, ymin, xmax, ymax)
                                                          for r in vipy.util.readcsv(os.path.join(labeldir, 'oidv6-train-annotations-bbox.csv'), ignoreheader=True)], keyfunc=lambda x: x[0])
    d_train_imageid_to_url = {r[0]:r[2] for r in vipy.util.readcsv(os.path.join(labeldir, 'oidv6-train-images-with-labels-with-rotation.csv'), ignoreheader=True)}    
    d_train_url_to_category = {r[2]:[d_label_to_category[c] for c in d_train_imageid_to_labels[r[0]]]
                               for r in vipy.util.readcsv(os.path.join(labeldir, 'oidv6-train-images-with-labels-with-rotation.csv'), ignoreheader=True) if r[0] in d_train_imageid_to_labels}

    #trainset = [(d_train_imageid_to_url[iid], [(d_label_to_object_category[o[1]], (o[2],o[3],o[4],o[5])) for o in obj if o[1] in d_label_to_object_category])
    #            for (iid, obj) in d_train_imageid_to_objects.items()]  # [(url, [(category, ulbr),...]), ...]    
    #loader = lambda r: vipy.image.Scene(url=r[0], tags=d_train_url_to_category[r[0]], objects=[vipy.object.Detection(category=c, ulbr=ulbr, normalized_coordinates=True) for (c,ulbr) in r[1]])
    #return vipy.dataset.Dataset(trainset, id='open_images_v7', loader=loader)
    
    imlist = [vipy.image.Scene(url=d_train_imageid_to_url[iid],
                               tags=d_train_url_to_category[d_train_imageid_to_url[iid]],
                               objects=[vipy.object.Detection(category=d_label_to_object_category[o[1]], ulbr=(o[2],o[3],o[4],o[5]), normalized_coordinates=True) for o in obj if o[1] in d_label_to_object_category])
              for (iid, obj) in d_train_imageid_to_objects.items()]  
    
    return vipy.dataset.Dataset(imlist, id='open_images_v7')




import os
import matplotlib

# Matplotlib Backend
# - Specify this environment variable VIPY_BACKEND
# - Headless operation: 'Agg'
# - Linux (X11-forwarded): 'TkAgg', 'GTK3Cairo', 'QtAgg'
# - Valid backends: https://matplotlib.org/stable/users/explain/backends.html
#
# Installing TkAgg
# sh> export VIPY_BACKEND='TkAgg'
# sh> sudo apt install python3.12-tk  # replace with your python version
# sh> pip install tk   # virtualenv
# >>> vipy.image.owl().show()

# Installing QtAgg
# sh> export VIPY_BACKEND='QtAgg'
# sh> sudo apt install python3-pyqt5   # replace with your python version
# sh> pip install pyqt5 opencv-python-headless 
# >>> vipy.image.owl().show()

if 'VIPY_BACKEND' in os.environ and 'DISPLAY' in os.environ:
    matplotlib.use(os.environ['VIPY_BACKEND'])  # 'Agg' is required unless there is a DISPLAY set

    
import importlib
BACKEND = importlib.import_module('vipy.gui.using_matplotlib')

try:
    import matplotlib.style as mplstyle
    mplstyle.use('fast')
except:
    pass


def figure(fignum=None):
    return BACKEND.figure(fignum)


def close(fignum=None):
    return BACKEND.close(fignum)


def closeall():
    return BACKEND.closeall()


def show(fignum):
    return BACKEND.show(fignum)


def noshow(fignum):
    return BACKEND.noshow(fignum)


def imshow(im, fignum=None, nowindow=False, timestamp=None, timestampcolor='black', timestampfacecolor=None, flush=False):
    """Show an image in the provided figure number"""
    if nowindow:
        noshow(fignum)
    h = BACKEND.imshow(im, fignum=fignum)
    if timestamp is not None:
        text(str(timestamp), 4, 15, fontsize=10, alpha=0.8 if timestampfacecolor is not None else 1.0, textfacealpha=0.6, facealpha=0.6, textcolor=timestampcolor, textfacecolor=timestampfacecolor)
    if not nowindow:
        show(fignum)
        if flush:
            BACKEND.imflush()
    return h


def imbbox(img, xmin, ymin, xmax, ymax, bboxcaption=None, fignum=None, nowindow=False):
    if nowindow:
        noshow(fignum)
    h = BACKEND.imshow(img, fignum=fignum)
    h = BACKEND.boundingbox(img, xmin, ymin, xmax, ymax, bboxcaption, fignum=h)
    if not nowindow:
        show(fignum)
        BACKEND.flush()
    return h


def imdetection(img, detlist, fignum=None, bboxcolor='green', facecolor='white', facealpha=0.5, do_caption=True, fontsize=10, textcolor='green', textfacecolor='white', textfacealpha=1.0, captionoffset=(0,0), nowindow=False, timestamp=None, timestampcolor='black', timestampfacecolor=None, timestampoffset=(0,0)):
    """Show a list of vipy.object.Detections overlayed on img.  Image must be RGB"""

    if nowindow:
        noshow(fignum)
    h = BACKEND.imdetection(img, detlist, fignum=fignum, bboxcolor=bboxcolor, do_caption=do_caption, facecolor=facecolor, facealpha=facealpha, fontsize=fontsize, textcolor=textcolor, captionoffset=captionoffset, textfacecolor=textfacecolor, textfacealpha=textfacealpha)
    if timestamp is not None:
        text(str(timestamp), 10+timestampoffset[0], 21+timestampoffset[1], fontsize=fontsize, textfacealpha=0.6, facealpha=0.6, alpha=0.8 if timestampfacecolor is not None else 1.0, textcolor=timestampcolor, textfacecolor=timestampfacecolor, pad=0.75)       
    if not nowindow:
        show(fignum)
        BACKEND.flush() if len(detlist)>0 else BACKEND.imflush()               
    return h

def imkeypoints(img, kplist, fignum=None, bordercolor='green', facecolor='white', facealpha=0.5, do_caption=True, fontsize=10, textcolor='green', textfacecolor='white', textfacealpha=1.0, captionoffset=(0,0), nowindow=False, timestamp=None, timestampcolor='black', timestampfacecolor=None, timestampoffset=(0,0)):

    if nowindow:
        noshow(fignum)
    h = BACKEND.imkeypoints(img, kplist, fignum=fignum, bordercolor=bordercolor, do_caption=do_caption, facecolor=facecolor, facealpha=facealpha, fontsize=fontsize, textcolor=textcolor, captionoffset=captionoffset, textfacecolor=textfacecolor, textfacealpha=textfacealpha)
    if timestamp is not None:
        text(str(timestamp), 10+timestampoffset[0], 21+timestampoffset[1], fontsize=fontsize, textfacealpha=0.6, facealpha=0.6, alpha=0.8 if timestampfacecolor is not None else 1.0, textcolor=timestampcolor, textfacecolor=timestampfacecolor, pad=0.75)       
    if not nowindow:
        show(fignum)
        BACKEND.flush() if len(kplist)>0 else BACKEND.imflush()               
    return h

def imobjects(img, objlist, fignum=None, bordercolor='green', facecolor='white', facealpha=0.5, do_caption=True, fontsize=10, textcolor='green', textfacecolor='white', textfacealpha=1.0, captionoffset=(0,0), nowindow=False, timestamp=None, timestampcolor='black', timestampfacecolor=None, timestampoffset=(0,0), timestamp_alpha=0.6, kp_alpha=0.7):
    if nowindow:
        noshow(fignum)
    h = BACKEND.imobjects(img, objlist, fignum=fignum, bordercolor=bordercolor, do_caption=do_caption, facecolor=facecolor, facealpha=facealpha, fontsize=fontsize, textcolor=textcolor, captionoffset=captionoffset, textfacecolor=textfacecolor, textfacealpha=textfacealpha, kp_alpha=kp_alpha)
    if timestamp is not None:
        text(str(timestamp), 10+timestampoffset[0], 21+timestampoffset[1], fontsize=fontsize, textfacealpha=0.6, facealpha=0.6, alpha=timestamp_alpha, textcolor=timestampcolor, textfacecolor=timestampfacecolor, pad=0.75)       
    if not nowindow:
        show(fignum)
        BACKEND.flush() if len(objlist)>0 else BACKEND.imflush()               
    return h


def impoints(img, objlist, fignum=None, bboxcolor='green', facecolor='white', facealpha=0.5, do_caption=True, fontsize=10, textcolor='green', textfacecolor='white', textfacealpha=1.0, captionoffset=(0,0), nowindow=False, timestamp=None, timestampcolor='black', timestampfacecolor=None, timestampoffset=(0,0)):
    pass

def frame(fmr, im=None, color='b.', caption=False, markersize=10, fignum=1):
    return BACKEND.frame(fr, im=im, color=color, caption=caption, markersize=markersize, fignum=fignum)


def imframe(img, fr, color='b', markersize=20, label=None, fignum=None):
    return BACKEND.imframe(img=img, fr=fr, color=color, markersize=markersize, label=label, fignum=fignum)


def savefig(filename=None, fignum=None, pad_inches=0, bbox_inches='tight', dpi=None, format=None):
    return BACKEND.savefig(filename, fignum, pad_inches=pad_inches, dpi=dpi, bbox_inches=bbox_inches, format=format)


def colorlist(theme=None):
    assert theme in [None, 'dark', 'light'], "invalid color theme '%s' - choose from [None, 'dark', 'light']" % theme
    return BACKEND.colorlist(theme=theme)


def text(caption, xmin, ymin, fignum=None, textcolor='black', textfacecolor=None, textfacealpha=1.0, fontsize=10, linewidth=3, facecolor='white', facealpha=0.5, alpha=1.0, pad=0.5):
    return BACKEND.text(caption, xmin, ymin, fignum, textcolor, textfacecolor, textfacealpha, fontsize, linewidth, facecolor, facealpha, alpha, pad=pad)
    

def array(img, mindim=512, figure=1):
    """Fast visualization of a numpy array img
        
    ```python
    vipy.show.array(np.random.rand(16,16,3))
    ```

    """
    from vipy.util import isnumpy
    from vipy.image import Image
    
    assert isnumpy(img)
    return Image(array=np.array(img).astype(np.float32), colorspace='float').mindim(mindim, interp='nearest').show(figure=figure)



import os
import vipy
from vipy.util import remkdir, isjpg
from vipy.image import ImageDetection
import vipy.downloader


URL = 'https://ethz.ch/content/dam/ethz/special-interest/itet/cvl/vision-dam/datasets/Dataset-information/ethz_shape_classes_v12.tgz'
SHA1 = 'ae9b8fad2d170e098e5126ea9181d0843505a84b'
SUBDIR = 'ETHZShapeClasses-V1.2'
LABELS = ['Applelogos','Bottles','Giraffes','Mugs','Swans']


class ETHZShapes(vipy.dataset.Dataset):
    def __init__(self, datadir=None, redownload=False):
        """ETHZShapes, provide a datadir='/path/to/store/ethzshapes' """

        datadir = tocache('ethzshapes') if datadir is None else datadir
        
        self._datadir = remkdir(datadir)

        if redownload or not os.path.exists(os.path.join(self._datadir, '.complete')):
            vipy.downloader.download_and_unpack(URL, self._datadir, sha1=SHA1)
        
        categorydir = LABELS
        imlist = []
        for (idx_category, category) in enumerate(categorydir):
            imdir = os.path.join(self._datadir, SUBDIR, category)
            for filename in os.listdir(imdir):
                if isjpg(filename) and not filename.startswith('.'):
                    # Write image
                    im = os.path.join(self._datadir, SUBDIR, category, filename)

                    # Write detections
                    gtfile = os.path.join(self._datadir, SUBDIR, category, os.path.splitext(os.path.basename(filename))[0] + '_' + category.lower() + '.groundtruth')
                    if not os.path.isfile(gtfile):
                        gtfile = os.path.join(self._datadir, SUBDIR, category, os.path.splitext(os.path.basename(filename))[0] + '_' + category.lower() + 's.groundtruth')  # plural hack
                    for line in open(gtfile,'r'):
                        if line.strip() == '':
                            continue
                        (xmin,ymin,xmax,ymax) = line.strip().split()
                        imlist.append( (im, category, xmin, ymin, xmax, ymax) )

        loader = lambda x: ImageDetection(filename=x[0], category=x[1], xmin=x[2], ymin=x[3], xmax=x[4], ymax=x[5])
        super().__init__(imlist, id='ethzshapes', loader=loader)

        open(os.path.join(self._datadir, '.complete'), 'a').close()



import os
import vipy

        
class PIP_175k(vipy.dataset.Dataset):
    URL = 'https://dl.dropboxusercontent.com/s/aqafx0t3k7691gc/pip_175k.tar.gz'
    MD5 = '2d8dce694e8a6056023c5232975297d9'
    def __init__(self, datadir):
        jsonfile = os.path.join(datadir, 'pip_175k', 'pip_175k.json')
        if not os.path.exists(jsonfile):

            if vipy.util.isinstalled('wget'):
                os.system('wget --no-check-certificate --continue --tries=32 -O %s %s ' % (os.path.join(datadir, vipy.util.filetail(PIP_175k.URL)), PIP_175k.URL))  # server fails many times, need smart continue
            else:
                vipy.downloader.download_and_unpack(PIP_175k.URL, datadir, md5=PIP_370k_stabilized.MD5)  # downloader fails many times ...

            if vipy.util.isinstalled('tar'):
                vipy.globals.log.info('extracting %s -> %s' % (os.path.join(datadir, vipy.util.filetail(PIP_175k.URL)), datadir))
                os.system('tar zxf %s --directory %s' % (os.path.join(datadir, vipy.util.filetail(PIP_175k.URL)), datadir))
            else:
                vipy.downloader.unpack(os.path.join(datadir, vipy.util.filetail(PIP_175k.URL)), datadir)

        super().__init__(vipy.load(jsonfile), id='pip_175k')

        
class PIP_370k_stabilized(vipy.dataset.Dataset):
    URL = 'https://dl.dropboxusercontent.com/s/fai9ontpmx4xv9i/pip_370k.tar.gz'
    MD5 = '2cf844fbc78fde1c125aa250e99db19f'
    def __init__(self, datadir):
        jsonfile = os.path.join(datadir, 'pip_370k', 'pip_370k.json')

        if not os.path.exists(jsonfile):        
            if vipy.util.isinstalled('wget'):
                os.system('wget --no-check-certificate --continue --tries=32 -O %s %s ' % (os.path.join(datadir, vipy.util.filetail(PIP_370k_stabilized.URL)), PIP_370k_stabilized.URL))  # server fails many times, need smart continue
            else:
                vipy.downloader.download_and_unpack(PIP_370k_stabilized.URL, datadir, md5=PIP_370k_stabilized.MD5)  # downloader fails many times ...

            if vipy.util.isinstalled('tar'):                
                vipy.globals.log.info('extracting %s -> %s' % (os.path.join(datadir, vipy.util.filetail(PIP_370k_stabilized.URL)), datadir))
                os.system('tar zxf %s --directory %s' % (os.path.join(datadir, vipy.util.filetail(PIP_370k_stabilized.URL)), datadir))
            else:
                vipy.downloader.unpack(os.path.join(datadir, vipy.util.filetail(PIP_370k_stabilized.URL)), datadir)
                
        super().__init__(vipy.load(jsonfile), id='pip_370k')



import os
import sqlite3
import csv
from vipy.util import remkdir
from vipy.image import ImageDetection
import numpy as np


SCHEMA = ['FILENAME', 'FACE_X', 'FACE_Y', 'FACE_WIDTH', 'FACE_HEIGHT',
          'LeftBrowLeftCorner_X', 'LeftBrowLeftCorner_Y', 'LeftBrowCenter_X', 'LeftBrowCenter_Y', 'LeftBrowRightCorner_X', 'LeftBrowRightCorner_Y',
          'RightBrowLeftCorner_X', 'RightBrowLeftCorner_Y', 'RightBrowCenter_X', 'RightBrowCenter_Y', 'RightBrowRightCorner_X', 'RightBrowRightCorner_Y',
          'LeftEyeLeftCorner_X', 'LeftEyeLeftCorner_Y', 'LeftEyeCenter_X', 'LeftEyeCenter_Y', 'LeftEyeRightCorner_X', 'LeftEyeRightCorner_Y',
          'RightEyeLeftCorner_X', 'RightEyeLeftCorner_Y', 'RightEyeCenter_X', 'RightEyeCenter_Y', 'RightEyeRightCorner_X', 'RightEyeRightCorner_Y',
          'LeftEar_X', 'LeftEar_Y', 'NoseLeft_X', 'NoseLeft_Y', 'NoseCenter_X', 'NoseCenter_Y', 'NoseRight_X', 'NoseRight_Y',
          'RightEar_X', 'RightEar_Y', 'MouthLeftCorner_X', 'MouthLeftCorner_Y', 'MouthCenter_X', 'MouthCenter_Y', 'MouthRightCorner_X', 'MouthRightCorner_Y',
          'ChinCenter_X', 'ChinCenter_Y']

LANDMARKS_3D = np.array([[-57.0899, 37.2398, 47.1156],
                         [-39.2328, 43.6173, 59.8098],
                         [-16.9114, 36.602, 65.2735],
                         [15.6142, 36.602, 65.2664],
                         [38.5733, 42.9796, 60.1435],
                         [55.7927, 37.2398, 49.5446],
                         [-50.0746, 20.6582, 48.3683],
                         [-34.1307, 21.9337, 56.6652],
                         [-19.4624, 18.7449, 53.8703],
                         [19.4407, 18.7449, 53.8993],
                         [33.4713, 21.2959, 56.124],
                         [46.8642, 21.2959, 50.6618],
                         [-76.2226, -26.5357, -12.6367],
                         [-17.5491, -23.9847, 68.4396],
                         [0.308036, -15.0561, 90.3836],
                         [18.8029, -25.898, 65.7138],
                         [78.1142, -27.1735, -16.2555],
                         [-28.3909, -53.3214, 58.766],
                         [-0.329719, -54.5969, 72.568],
                         [28.3693, -53.9592, 58.5929],
                         [-0.967474, -96.051, 63.4542]])


class AFLW(object):
    def __init__(self, datadir):
        self.datadir = remkdir(datadir)
        if not os.path.isdir(os.path.join(self.datadir)):
            raise ValueError('Download AFLW dataset manually to "%s" ' % self.datadir)

    def __repr__(self):
        return str('<vipy.data.aflw: %s>' % self.datadir)

    def dataset(self):
        csvfile = os.path.join(self.datadir, 'aflw.csv')
        with open(csvfile, 'r') as f:
            for x in f.readline().split(','):
                if x[0][0] != '#':
                    im = ImageDetection(filename=os.path.join(self.datadir, x[0]), category='face',
                                        xmin=float(x[1]) if len(x[1]) > 0 else float('nan'),
                                        ymin=float(x[2]) if len(x[2]) > 0 else float('nan'),
                                        xmax=float(x[1]) + float(x[3]) if ((len(x[1]) > 0) and (len(x[3]) > 0)) else float('nan'),
                                        ymax=float(x[2]) + float(x[4]) if ((len(x[2]) > 0) and (len(x[4]) > 0)) else float('nan'),
                                        attributes={k:v for (k,v) in zip(SCHEMA,x)})  # Parse row
                    yield im

    def export(self):
        """Export sqlite database file to aflw.csv"""
        dbfile = os.path.join(self.datadir, 'data', 'aflw.sqlite')
        db = sqlite3.connect(dbfile)
        cursor = db.cursor()

        outfile = os.path.join(self.datadir, 'aflw.csv')
        with open(outfile, 'wb') as csvfile:
            f = csv.writer(csvfile, delimiter=',', quotechar='|', quoting=csv.QUOTE_MINIMAL)
            f.writerow([s if k > 0 else '#' + s for (k,s) in enumerate(SCHEMA)])  # comment first line
            faceidQuery = cursor.execute('SELECT face_id FROM Faces').fetchall()
            for faceid in faceidQuery:
                faceid = faceid[0]
                fileidQuery = cursor.execute('SELECT file_id FROM Faces WHERE face_id = "%s"' % str(faceid))
                fileid = str(fileidQuery.fetchone()[0])

                imgDataQuery = cursor.execute('SELECT db_id,filepath,width,height FROM FaceImages where file_id = "%s"' % str(fileid))
                imgData = imgDataQuery.fetchall()

                facerect = cursor.execute('SELECT x,y,w,h FROM FaceRect WHERE face_id = "%s"' % str(faceid)).fetchone()

                ptsQuery = cursor.execute('SELECT descr,FeatureCoords.x,FeatureCoords.y FROM FeatureCoords,FeatureCoordTypes WHERE face_id = "%s" AND FeatureCoords.feature_id = FeatureCoordTypes.feature_id' % (str(faceid))).fetchall()
                annoDict = {}
                for pts in ptsQuery:
                    annoDict['%s_X' % str(pts[0])] = pts[1]
                    annoDict['%s_Y' % str(pts[0])] = pts[2]

                annoDict['FILENAME'] = os.path.join('data','flickr', imgData[0][1])
                annoDict['FACE_X'] = facerect[0]
                annoDict['FACE_Y'] = facerect[1]
                annoDict['FACE_WIDTH'] = facerect[2]
                annoDict['FACE_HEIGHT'] = facerect[3]

                row = [annoDict[key] if key in annoDict.keys() else '' for key in SCHEMA]
                print('[vipy.data.aflw]: exporting %d points for face "%s" ' % (len(ptsQuery), faceid))
                f.writerow(row)

        db.close()
        return self


def landmarks(im):
    """Return 21x2 frame array of landmark positions in 1-21 order, NaN if occluded"""
    return np.float32(np.array([im.attributes[key] if len(im.attributes[key]) > 0 else np.float32('nan') for key in SCHEMA[5:]])).reshape(21, 2)


def eyes_nose_chin(self, im):
    """Return 4x2 frame array of left eye, right eye nose chin"""
    fr = landmarks(im)
    return fr[[8 - 1, 11 - 1, 15 - 1, 21 - 1],:]  # left eye center, right eye center, nose center  (AFLW annotation, 1-indexed)



import vipy
import numpy as np
from io import BytesIO
import PIL
import random
import math


vipy.util.try_import('torchvision.transforms', 'torchvision'); import torchvision.transforms
vipy.util.try_import('torchvision.transforms.v2', 'torchvision'); import torchvision.transforms.v2
vipy.util.try_import('skimage', 'scikit-image'); import skimage.transform
vipy.util.try_import('cv2', 'opencv-python'); import cv2


# GEOMETRIC
def left_shear(im, s, border='zero'):
    assert isinstance(im, vipy.image.Image)
    assert s >= 0
    
    T = vipy.geometry.affine_transform(txy=(-im.width()//2, -im.height()/2))
    A = np.dot(vipy.geometry.affine_transform(ky=s), T)
    A = np.dot(np.linalg.inv(T), A)
    return im.load().affine_transform(A, border=border).mat2gray(0,255)

def right_shear(im, s, border='zero'):
    assert isinstance(im, vipy.image.Image)
    assert s >= 0
    
    T = vipy.geometry.affine_transform(txy=(-im.width()//2, -im.height()/2))
    A = np.dot(vipy.geometry.affine_transform(ky=-s), T)
    A = np.dot(np.linalg.inv(T), A)
    return im.load().affine_transform(A, border=border).mat2gray(0,255)

def rotate(im, rad=None, deg=None, border='zero'):
    assert isinstance(im, vipy.image.Image)
    assert rad is not None or deg is not None
    
    r = rad if rad is not None else (deg*np.pi/180.0)
    T = vipy.geometry.affine_transform(txy=(-im.width()//2, -im.height()/2))
    A = np.dot(vipy.geometry.affine_transform(r=r), T)
    A = np.dot(np.linalg.inv(T), A)
    return im.load().affine_transform(A, border=border).mat2gray(0,255)

def barrel(im):
    assert isinstance(im, vipy.image.Image)    
    (width, height) = (im.width(), im.height())
    distCoeff = np.zeros((4,1),np.float64)
    
    k1 = 0.1 + 0.05*np.random.randn()
    k2 = 0.1 + 0.05*np.random.randn()
    p1 = 0.1 + 0.05*np.random.randn()
    p2 = 0.1 + 0.05*np.random.randn()
    
    distCoeff[0,0] = k1;
    distCoeff[1,0] = k2;
    distCoeff[2,0] = p1;
    distCoeff[3,0] = p2;
    
    # assume unit matrix for camera
    cam = np.eye(3,dtype=np.float32)

    cam[0,2] = width/2.0  # define center x
    cam[1,2] = height/2.0 # define center y
    cam[0,0] = 600.        # define focal length x
    cam[1,1] = 600.        # define focal length y

    # here the undistortion will be computed
    dst = cv2.undistort(im.load().array(),cam,distCoeff)

    return im.array(dst)

def left_swirl(im, center=None, strength=2):
    assert isinstance(im, vipy.image.Image)        
    return im.array(np.array(skimage.transform.swirl(im.load().array(), rotation=0, strength=strength, center=center, radius=im.maxdim())).astype(np.float32)).mat2gray(0,1)

def right_swirl(im, strength=2, center=None):
    assert isinstance(im, vipy.image.Image)            
    return im.array(np.array(skimage.transform.swirl(im.load().fliplr().array(), rotation=0, center=center, strength=strength, radius=im.maxdim())).astype(np.float32)).mat2gray(0,1).fliplr()

def horizontal_mirror(im):
    assert isinstance(im, vipy.image.Image)            
    img = im.load().array()
    (i,j) = (im.width()//2 if vipy.math.iseven(im.width()) else (im.width()//2)+1, im.width()//2)    
    img[:,i:] = np.fliplr(img[:,0:j])
    return im.array(img)

def vertical_mirror(im):
    assert isinstance(im, vipy.image.Image)            
    img = im.load().array()
    (i,j) = (im.height()//2 if vipy.math.iseven(im.height()) else (im.height()//2)+1, im.height()//2)
    img[i:,:] = np.flipud(img[0:j, :])
    return im.array(img)

def vertical_motion_blur(im, kernel_size):
    assert isinstance(im, vipy.image.Image)
    kernel_v = np.zeros((kernel_size, kernel_size) )  
    kernel_v[:, int((kernel_size - 1)/2)] = np.ones(kernel_size)  
    kernel_v /= kernel_size
    return im.load().map(lambda img: cv2.filter2D(img, -1, kernel_v)).mat2gray(0,255).colorspace('float')

def horizontal_motion_blur(im, kernel_size):
    assert isinstance(im, vipy.image.Image)    
    kernel_h = np.zeros((kernel_size, kernel_size) )  
    kernel_h[int((kernel_size - 1)/2), :] = np.ones(kernel_size)
    kernel_h /= kernel_size
    return im.load().map(lambda img: cv2.filter2D(img, -1, kernel_h)).mat2gray(0,255).colorspace('float')

def ghost(im, txr=5, tyr=5, txg=-5, tyg=-5, txb=5, tyb=-5):
    assert isinstance(im, vipy.image.Image)
    (imc,imz) = (im.clone().load(), im.clone().load().zeros())
    img = imz.numpy()
    img[:,:,0] = translate(imc.red(), txr, tyr).load().array()
    img[:,:,1] = translate(imc.green(), txg, tyg).load().array()
    img[:,:,2] = translate(imc.blue(), txb, tyb).load().array()
    return imz

def crop(im, s=0.8):
    return im.clone().crop(vipy.geometry.BoundingBox(xmin=random.randint(0, int((1-s)*im.width())), ymin=random.randint(0, int((1-s)*im.height())), width=math.ceil(s*im.width()), height=math.ceil(s*im.height()))).resize(height=im.height(), width=im.width())

def fliplr(im):
    return im.load().fliplr()

def flipud(im):
    return im.load().flipud()

def rot90cw(im):
    return im.load().rot90cw()

def rot90ccw(im):
    return im.load().rot90ccw()

def translate(im, dx, dy): 
    """Translate by (dx,dy) pixels, with zero border handling"""
    assert isinstance(im, vipy.image.Image)
    return im.load().padcrop(im.imagebox().translate(dx,dy))

def isotropic_scale(im, s, border='zero'):
    assert isinstance(im, vipy.image.Image)
    assert s > 0

    (W,H) = (im.width(), im.height())
    return im.load().rescale(s).padcrop(vipy.geometry.BoundingBox(centroid=(im.width()//2, im.height()//2), width=W, height=H))

def zoom(im, s, zx=None, zy=None, border='zero'):
    assert isinstance(im, vipy.image.Image)
    assert s > 0

    (zx, zy) = ((-im.width()/2) if zx is None else -zx, (-im.height()/2) if zy is None else -zy)    
    return isotropic_scale(translate(im, zx, zy), s)


# <PHOTOMETRIC>
def blur(im, sigma):
    assert isinstance(im, vipy.image.Image)
    imblur = im.rgb().blur(sigma=sigma)
    return imblur.greyscale().colorspace_like(im) if im.channels() == 1 else imblur.colorspace_like(im)

def salt_and_pepper(im, p):
    assert isinstance(im, vipy.image.Image)    
    assert p >= 0 and p <= 1
    
    W = np.array(np.random.rand( im.height(), im.width(), 1 ) > p).astype(np.float32)
    B = np.array(np.random.rand( im.height(), im.width(), 1 ) > p).astype(np.float32)    
    return im.load().float().mat2gray().map(lambda x: np.multiply(x, W) + 1-W).map(lambda x: np.multiply(x, B)).colorspace_like(im)

def jpeg_compression(im, quality):
    assert isinstance(im, vipy.image.Image)
    assert quality >= 0 and quality <= 95
    
    out = BytesIO()
    im.load().pil().save(out, format='jpeg', quality=quality)
    out.seek(0)
    return im.load().array(np.array(PIL.Image.open(out))).mat2gray(0,255)

def bit_depth(im, d):
    assert d>=1 and d<=8    
    return im.array(np.array(PIL.ImageOps.posterize(im.load().pil(), d)))

def solarize(im, t):
    assert t>=0 and t<=255    
    return im.array(np.array(PIL.ImageOps.solarize(im.load().pil(), t)))

def permute_color_channels(im, order):
    assert sorted(order) == [0,1,2]
    return im.array(np.array(im.load().rgb())[:,:,order])

def greyscale(im):
    assert isinstance(im, vipy.image.Image)    
    return im.clone().load().greyscale().colorspace_like(im)

def bgr(im):
    assert isinstance(im, vipy.image.Image)    
    return im.array(im.bgr().load().array()).colorspace('rgb')

def hot(im):
    assert isinstance(im, vipy.image.Image)    
    return im.hot().load().rgb()

def rainbow(im):
    assert isinstance(im, vipy.image.Image)    
    return im.rainbow().load().rgb()

def saturate(im, low, high):
    assert isinstance(im, vipy.image.Image)    
    return im.saturate(low, high)

def colorjitter(im):
    transform = torchvision.transforms.ColorJitter(brightness=(0.5,1.5),contrast=(1),saturation=(0.5,1.5),hue=(-0.1,0.1))
    return im.array(np.array(transform(im.pil())))

def sharpness(im, f):
    return im.array(np.array(torchvision.transforms.v2.functional.adjust_sharpness(im.load().torch('CHW'), f).permute(1,2,0)))

def gamma(im, f):
    return im.array(np.array(torchvision.transforms.v2.functional.adjust_gamma(im.load().torch('CHW'), f).permute(1,2,0)))

def autocontrast(im):
    return im.mat2gray().rgb()

def edge(im):
    assert isinstance(im, vipy.image.Image)    
    return im.array(np.array(im.luminance().pil().filter(PIL.ImageFilter.FIND_EDGES))).colorspace('lum').colorspace_like(im)

def emboss(im): 
    assert isinstance(im, vipy.image.Image)   
    return im.array(np.array(im.luminance().pil().filter(PIL.ImageFilter.EMBOSS))).colorspace('lum').colorspace_like(im)

def darken(im, g): 
    assert isinstance(im, vipy.image.Image)   
    return im.gain(g).mat2gray(0, 255).colorspace('float')

def negative(im):
    assert isinstance(im, vipy.image.Image)    
    return im.load().array(1-im.load().mat2gray(0,255).array()).colorspace('float')

def scan_lines(im):
    assert isinstance(im, vipy.image.Image)
    imc = im.load()
    imc._array[::2, :] = 0
    return imc

def additive_gaussian_noise(im, sigma=0.1):
    assert isinstance(im, vipy.image.Image)            
    return im.float().map(lambda img: img + max(sigma*im.maxpixel(), sigma)*np.random.randn(img.shape[0], img.shape[1], im.channels()).astype(np.float32)).saturate(0,255).mat2gray(0,255).colorspace('float')



class Noise():
    def __init__(self, magnitude=0.25, provenance=False, register=None, deregister=['hot','rainbow','vertical_mirror','flipud','rot90cw','rot90ccw']):

        n = magnitude
        (shear, dx, dy, rot, z, border, sigma, gain) = (n*0.25, n*0.15, n*0.15, n*25, n*0.2, 'replicate', 1.0, 0.5)
        
        self._provenance = provenance        
        self._registry = {'blur': lambda im: blur(im, sigma=n*random.uniform(im.mindim()*0.01, im.mindim()*0.05)),
                          'salt_and_pepper': lambda im: salt_and_pepper(im, n*0.1*random.random()),
                          'jpeg_compression': lambda im: jpeg_compression(im, random.randint(50,95-int(10*n))),
                          'greyscale': lambda im: greyscale(im),
                          'bgr': lambda im: bgr(im),
                          'hot': lambda im: hot(im),
                          'rainbow': lambda im: rainbow(im),
                          'saturate': lambda im: saturate(im, random.randint(0, 64), random.randint(255-64, 255)),
                          'edge': lambda im: edge(im),
                          'emboss': lambda im: emboss(im),
                          'darken': lambda im: darken(im, 1-n*random.random()),
                          'negative': lambda im: negative(im),
                          'scan_lines': lambda im: scan_lines(im),
                          'additive_gaussian_noise': lambda im: additive_gaussian_noise(im, np.random.uniform(n*0.01, n*0.5)),
                          'translate': lambda im, dx=dx, dy=dy, b=border: translate(im, np.random.uniform(-dx,dx), np.random.uniform(-dy,dy)),                          
                          'horizontal_motion_blur': lambda im: horizontal_motion_blur(im, random.randint(math.ceil(im.mindim()*0.05*n), math.ceil(im.mindim()*0.15*n))),
                          'vertical_motion_blur': lambda im: vertical_motion_blur(im, random.randint(math.ceil(im.mindim()*0.05*n), math.ceil(im.mindim()*0.15*n))),                          
                          'barrel': lambda im: barrel(im),
                          'left_swirl': lambda im: left_swirl(im, center=(random.randint(0, im.mindim()), random.randint(0,im.mindim())), strength=n),
                          'right_swirl': lambda im: right_swirl(im, center=(random.randint(0, im.mindim()), random.randint(0, im.mindim())), strength=n),
                          'horizontal_mirror': lambda im: horizontal_mirror(im),
                          'vertical_mirror': lambda im: vertical_mirror(im),
                          'left_shear': lambda im, s=shear, b=border: left_shear(im, np.random.uniform(0,s), border=b),
                          'right_shear': lambda im, s=shear, b=border: right_shear(im, np.random.uniform(0,s), border=b),
                          'rotate': lambda im, deg=rot, b=border: rotate(im, deg=np.random.uniform(-deg,deg), border=b),
                          'isotropic_scale': lambda im, z=z, b=border: isotropic_scale(im, np.random.uniform(1.0-z, 1.0+z)),
                          'ghost': lambda im: ghost(im,
                                                    random.randint(-int(im.mindim()*0.05*n),int(im.mindim()*0.05*n)),
                                                    random.randint(-int(im.mindim()*0.05*n),int(im.mindim()*0.05*n)),
                                                    random.randint(-int(im.mindim()*0.05*n),int(im.mindim()*0.05*n)),
                                                    random.randint(-int(im.mindim()*0.05*n),int(im.mindim()*0.05*n)),
                                                    random.randint(-int(im.mindim()*0.05*n),int(im.mindim()*0.05*n)),
                                                    random.randint(-int(im.mindim()*0.05*n),int(im.mindim()*0.05*n))),
                          'bit_depth': lambda im: bit_depth(im, random.randint(1,7)),
                          'permute_color_channels': lambda im: permute_color_channels(im, random.sample([[0,2,1],[1,0,2],[1,2,0],[2,0,1],[2,1,0]], 1)[0]),
                          'fliplr': lambda im: fliplr(im),
                          'flipud': lambda im: flipud(im),
                          'rot90cw': lambda im: rot90cw(im),
                          'rot90ccw': lambda im: rot90ccw(im),
                          'solarize': lambda im: solarize(im, random.randint(32, 255-32)),
                          'colorjitter': lambda im: colorjitter(im),
                          'autocontrast': lambda im: autocontrast(im),
                          'sharpness': lambda im: sharpness(im, 10*random.random()),
                          'gamma': lambda im: gamma(im, random.random()+0.5),
                          'crop': lambda im: crop(im, 1-0.2*random.random()),
                          }

        if deregister is not None:
            self._registry = {k:v for (k,v) in self._registry.items() if k not in deregister}                    
        if register is not None:
            self._registry = {k:v for (k,v) in self._registry.items() if k in register}        

    def transformations(self):
        return list(self._registry.keys())

    def random_transformation(self, im):
        return self.transform(im, vipy.util.takeone(self.transformations()))
        
    def transform(self, im, transform):
        assert isinstance(im, vipy.image.Image)
        assert transform in self.transformations()

        imd = self._registry[transform](im.clone().load())
        if self._provenance:
            imd.setattribute('vipy.noise', transform)
        return imd

    def montage(self, im):
        return vipy.visualize.montage([self.transform(im.clone().centersquare(), k) for k in self.transformations()], 256, 256)
    
    def __call__(self, im):
        assert isinstance(im, vipy.image.Image) or all(isinstance(x, vipy.image.Image) for x in im), "singletons or batches required"
        return self.random_transformation(im) if isinstance(im, vipy.image.Image) else tuple([self.random_transformation(x) for x in im])


class RandomCrop(Noise):
    def __init__(self, magnitude=1, provenance=False):
        super().__init__(magnitude=magnitude, provenance=provenance, register=['translate', 'isotropic_scale', 'crop'])
    
class Geometric(Noise):
    def __init__(self, magnitude=0.25, provenance=False):
        super().__init__(magnitude=magnitude, provenance=provenance,
                         register=['translate', 'horizontal_motion_blur', 'vertical_motion_blur', 'barrel',
                                   'left_swirl', 'right_swirl','horizontal_mirror','left_shear','right_shear',
                                   'rotate','isotropic_scale','fliplr','crop'])
class Photometric(Noise):
    def __init__(self, magnitude=0.25, provenance=False):
        super().__init__(magnitude=magnitude, provenance=provenance,
                         register=['blur', 'salt_and_pepper', 'jpeg_compression', 'greyscale', 'bgr', 'hot','saturate','edge','emboss',
                                   'darken','negative','scan_lines','additive_gaussian_noise','bit_depth','permute_color_channels','solarize',
                                   'colorjitter','autocontrast', 'sharpness', 'gamma', 'ghost'])
class Perturbation(Noise):
    def __init__(self, magnitude=0.5, provenance=False):
        super().__init__(magnitude=magnitude, provenance=provenance, register=['translate', 'isotropic_scale', 'crop', 'blur', 'additive_gaussian_noise']) 
        

geometric = Geometric(provenance=True)
photometric = Photometric(provenance=True)
perturbation = Perturbation(provenance=True)
randomcrop = RandomCrop(provenance=True)




import os
import vipy
from vipy.util import readcsv, remkdir, filepath, islist, filetail, filebase, filefull, tocache, isinstalled
from vipy.image import ImageDetection, ImageCategory
import scipy.io
import numpy as np
from vipy.globals import log

try:
    import lxml.etree as ET  # faster (optional)
except:
    import xml.etree.ElementTree as ET  # slower (default)

    
URLS_2014 = ['https://image-net.org/data/ILSVRC/2014/ILSVRC2014_DET_train.tar',
             'https://image-net.org/data/ILSVRC/2013/ILSVRC2013_DET_val.tar',
             'https://image-net.org/data/ILSVRC/2014/ILSVRC2014_DET_bbox_train.tgz',
             'https://image-net.org/data/ILSVRC/2013/ILSVRC2013_DET_bbox_val.tgz',
             'https://image-net.org/data/ILSVRC/2012/ILSVRC2012_bbox_train_v2.tar.gz',
             'https://image-net.org/data/ILSVRC/2012/ILSVRC2012_bbox_val_v3.tgz',
             'https://image-net.org/data/ILSVRC/2013/ILSVRC2013_DET_test.tar',
             'https://image-net.org/data/bboxes_annotations.tar.gz']
             
URLS_2012 = ['https://image-net.org/data/ILSVRC/2012/ILSVRC2012_img_train.tar',
             'https://image-net.org/data/ILSVRC/2012/ILSVRC2012_img_val.tar',             
             'https://image-net.org/data/ILSVRC/2012/ILSVRC2012_bbox_val_v3.tgz',
             'https://image-net.org/data/ILSVRC/2012/ILSVRC2012_bbox_train_v2.tar.gz',
             'https://image-net.org/data/ILSVRC/2012/ILSVRC2012_img_train_t3.tar',
             'https://image-net.org/data/ILSVRC/2012/ILSVRC2012_bbox_train_dogs.tar.gz',
             'https://image-net.org/data/ILSVRC/2012/ILSVRC2012_devkit_t12.tar.gz']

URL_SYNSET = 'https://raw.githubusercontent.com/torch/tutorials/master/7_imagenet_classification/synset_words.txt'
IMAGENET21K_RESIZED_URL = 'https://image-net.org/data/imagenet21k_resized.tar.gz'
IMAGENET21K_URL = 'https://image-net.org/data/winter21_whole.tar.gz'
IMAGENET21K_WORDNET_ID = 'https://storage.googleapis.com/bit_models/imagenet21k_wordnet_ids.txt'
IMAGENET21K_WORDNET_LEMMAS = 'https://storage.googleapis.com/bit_models/imagenet21k_wordnet_lemmas.txt'
IMAGENET21K_MD5 = 'ab313ce03179fd803a401b02c651c0a2'

IMAGENET_FACES = 'https://image-net.org/data/face_annotations_ILSVRC.json'


class Imagenet2012():
    """Imagenet2012 requires login at https://image-net.org from the same IP address as the download, and agreeing to the ImageNet terms: https://image-net.org/download-images.php#term"""
    def __init__(self, datadir=None, redownload=False):
        datadir = tocache('imagenet2012') if datadir is None else datadir
        
        self._datadir = remkdir(datadir)

        if redownload or not os.path.exists(os.path.join(self._datadir, '.complete')):        
            for url in URLS_2012:
                vipy.downloader.download(url, os.path.join(self._datadir, filetail(url)))

            for url in URLS_2012:
                vipy.downloader.unpack(os.path.join(self._datadir, filetail(url)), remkdir(os.path.join(self._datadir, filebase(url))))

            for f in vipy.util.findtar(os.path.join(self._datadir, 'ILSVRC2012_img_train')):
                vipy.downloader.unpack(f, filefull(f))

            for f in vipy.util.findtar(os.path.join(self._datadir, 'ILSVRC2012_img_train_v3')):
                vipy.downloader.unpack(f, filefull(f))
               
            vipy.downloader.download(URL_SYNSET, os.path.join(self._datadir, 'synset_words.txt'))
            open(os.path.join(self._datadir, '.complete'), 'a').close()        
            
        self._synset_to_categorylist = {x.split(' ',1)[0]:[y.lstrip().rstrip() for y in x.split(' ', 1)[1].split(',')] for x in vipy.util.readtxt(os.path.join(self._datadir, 'synset_words.txt'))}            

        metadata = scipy.io.loadmat(os.path.join(self._datadir, 'ILSVRC2012_devkit_t12/ILSVRC2012_devkit_t12/data/meta.mat'), struct_as_record=False)
        synsets = np.squeeze(metadata['synsets'])
        ids = np.squeeze(np.array([s.ILSVRC2012_ID for s in synsets]))
        wnids = np.squeeze(np.array([s.WNID for s in synsets]))
        words = np.squeeze(np.array([s.words for s in synsets]))
            
        self._wnid_to_categorylist = {wnid:[c.strip() for c in category.split(',')] for (wnid,category) in zip(wnids, words)}
                                    
        
    def synset_to_category(self, s=None):
        return self._wnid_to_categorylist if s is None else self._wnid_to_categorylist[s]
    
    def classification_trainset(self):
        """ImageNet2012 Classification, trainset"""
        imgfiles = vipy.util.findimages(os.path.join(self._datadir, 'ILSVRC2012_img_train'))  # slow-ish, may be better to cache
        imlist = ([vipy.image.TaggedImage(filename=f, attributes={'wordnet_id':f.rsplit('/',2)[-2]}, tags=self._wnid_to_categorylist[f.rsplit('/',2)[-2]]) for f in imgfiles] if os.name == 'posix' else
                  [vipy.image.TaggedImage(filename=f, attributes={'wordnet_id':os.path.basename(os.path.dirname(f))}, tags=self._wnid_to_categorylist[os.path.basename(os.path.dirname(f))]) for f in imgfiles])
        
        return vipy.dataset.Dataset(imlist, 'imagenet2012_classification:train')
        
    def classification_valset(self):
        """ImageNet2012 Classification, valset"""        
        imlist = []
        imgfiles = vipy.util.findimages(os.path.join(self._datadir, 'ILSVRC2012_img_val'))  # slow-ish, may be better to cache
                    
        # ground truth is imagenet synset index 1-1000
        gt = vipy.util.readtxt(os.path.join(self._datadir, 'ILSVRC2012_devkit_t12', 'ILSVRC2012_devkit_t12', 'data', 'ILSVRC2012_validation_ground_truth.txt'))
        for (f,y) in zip(sorted(imgfiles), gt):
            imlist.append( (f, y) )

        # Index mapping is in mat file (yuck)
        synsets = self.synset_to_category()
        d_idx_to_category = {str(k):self.synset_to_category(r[0][1][0]) for (k,r) in enumerate(scipy.io.loadmat(os.path.join(self._datadir, 'ILSVRC2012_devkit_t12/ILSVRC2012_devkit_t12/data/meta.mat'))['synsets'], start=1) if r[0][1][0] in synsets}

        imlist = [vipy.image.TaggedImage(filename=f, attributes={'synset_index':y}, tags=d_idx_to_category[y]) for (f,y) in imlist] 
        return vipy.dataset.Dataset(imlist, 'imagenet2012_classification:val')

    def faces(self):
        """Return all annotated faces in 2012 train and val sets:
        https://image-net.org/face-obfuscation/
        """

        cachefile = os.path.join(self._datadir, 'faces.json')
        if not os.path.exists(cachefile):        
            if not os.path.exists(os.path.join(self._datadir, 'face_annotations_ILSVRC.json')):
                vipy.downloader.download(IMAGENET_FACES, os.path.join(self._datadir, filetail(IMAGENET_FACES)))

            faces = {vipy.util.filetail(d['url']):d['bboxes'] for d in vipy.util.readjson(os.path.join(self._datadir, 'face_annotations_ILSVRC.json'))}
            imgfiles = vipy.util.findimages(os.path.join(self._datadir, 'ILSVRC2012_img_train'))  # slow-ish, may be better to cache
            imgfiles += vipy.util.findimages(os.path.join(self._datadir, 'ILSVRC2012_img_val'))  # slow-ish, may be better to cache
            
            imlist = [vipy.image.Scene(filename=f).objects([vipy.object.Detection(category='face', xmin=o['x0'], ymin=o['y0'], xmax=o['x1'], ymax=o['y1']) for o in (faces[filetail(f)] if filetail(f) in faces else [])]) for f in imgfiles]
            vipy.save(imlist, cachefile)
        else:
            try:
                imlist = vipy.load(cachefile)
            except:
                if os.path.exists(cachefile):
                    os.remove(cachefile)  # force recache on failure
                return self.faces()
            
        return vipy.dataset.Dataset([im for im in imlist if im.num_objects()>0], 'imagenet2012_faces:train')
        
    def localization_trainset(self):
        """ImageNet2012 localization, imageset = {train, val}, this takes a long time to read the XML files, load and cache"""

        cachefile = os.path.join(self._datadir, 'localization_trainset.json')
        if not os.path.exists(cachefile):
            log.warning('Initial parsing of XML files for imagenet-localization takes a long time... Consider "pip install lxml ujson" to speed this up')
        
            imlist = []
            classification = self.classification_trainset()
            synsets = self.synset_to_category()
            for f in classification._ds:                
                xmlfile = '%s.xml' % filefull(f.replace('ILSVRC2012_img_train', 'ILSVRC2012_bbox_train_v2'))
                im = vipy.image.Scene(filename=f)
                if os.path.exists(xmlfile):
                    root = ET.parse(xmlfile).getroot()  
                    for obj in root.findall("object"):
                        name = obj.findtext("name")
                        b = obj.find("bndbox")
                        xmin = int(b.findtext("xmin"))
                        ymin = int(b.findtext("ymin"))
                        xmax = int(b.findtext("xmax"))
                        ymax = int(b.findtext("ymax"))
                        im.add_object(vipy.object.Detection(tags=synsets[name] if name in synsets else [name], xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax))
                if im.num_objects()>0:
                    imlist.append(im)                
            vipy.save(imlist, cachefile)  # cached
            
        else:
            try:
                imlist = vipy.load(cachefile)
            except:
                if os.path.exists(cachefile):
                    os.remove(cachefile)  # force recache on failure
                return self.localization_trainset()  
                               
        return vipy.dataset.Dataset(imlist, 'imagenet2012_localization:train')
    
                
class Imagenet21K_Resized(vipy.dataset.Dataset):
    """Imagenet21K_Resized requires login at https://image-net.org from the same IP address as the download, and agreeing to the ImageNet terms: https://image-net.org/download-images.php#term    
       https://image-net.org/download-images.php, imagenet-21K 2021 release ("squish" resized)"""
    def __init__(self, datadir=None, aslemma=True, redownload=False, recache=False):

        datadir = tocache('imagenet21k_resized') if datadir is None else datadir
        
        self._datadir = vipy.util.remkdir(datadir)
        
        if redownload or not os.path.exists(os.path.join(self._datadir, '.complete')):
            vipy.globals.log.info('[vipy.data.imagenet]: downloading Imagenet-21K resized to "%s"' % self._datadir)            
            vipy.downloader.download_and_unpack(IMAGENET21K_RESIZED_URL, self._datadir, sha1=None)
            vipy.downloader.download(IMAGENET21K_WORDNET_ID, os.path.join(self._datadir, 'imagenet21k_wordnet_ids.txt'))
            vipy.downloader.download(IMAGENET21K_WORDNET_LEMMAS, os.path.join(self._datadir, 'imagenet21k_wordnet_lemmas.txt'))

            open(os.path.join(self._datadir, '.complete'), 'a').close()
        
        # Class names: https://github.com/google-research/big_transfer/issues/7
        self._synset_to_categorylist = {x:[y.rstrip().lstrip() for y in lemma.split(',')] for (x,lemma) in zip(vipy.util.readtxt(os.path.join(self._datadir, 'imagenet21k_wordnet_ids.txt')), vipy.util.readtxt(os.path.join(self._datadir, 'imagenet21k_wordnet_lemmas.txt')))}

        cachefile = os.path.join(self._datadir, '.imlist.txt')
        if recache and os.path.exists(cachefile):
            os.remove(cachefile)
            
        f_category = lambda c, synset_to_categorylist=self._synset_to_categorylist, aslemma=aslemma: synset_to_categorylist[c][0] if aslemma else c        
        imlist = vipy.util.findimages(os.path.join(datadir, 'imagenet21k_resized')) if not os.path.exists(cachefile) else [os.path.join(self._datadir, f) for f in vipy.util.readlist(cachefile)]
        loader = lambda f, f_category=f_category: vipy.image.TaggegdImage(filename=f,
                                                                          attributes={'wordnet_id':vipy.util.filebase(vipy.util.filepath(f))},
                                                                          tags=f_category(vipy.util.filebase(vipy.util.filepath(f))))
        super().__init__(imlist, id='imagenet21k_resized', loader=loader)

        if not os.path.exists(cachefile):
            vipy.util.writelist([f.replace(self._datadir + '/', '') for f in imlist], cachefile)  # cache me for faster loading instead of walking the directory tree        

        
    def synset_to_category(self, s=None):
        return self._synset_to_categorylist if s is None else self._synset_to_categorylist[s]
        

class Imagenet21K(vipy.dataset.Dataset):
    """Imagenet21K requires login at https://image-net.org from the same IP address as the download, and agreeing to the ImageNet terms: https://image-net.org/download-images.php#term    
       imagenet-21K 2021 winter release
    """
    def __init__(self, datadir=None, aslemma=True, redownload=False, recache=False):

        datadir = tocache('imagenet21k') if datadir is None else datadir
        
        self._datadir = vipy.util.remkdir(datadir)
        
        if redownload or not os.path.exists(os.path.join(self._datadir, '.complete')):
            if isinstalled('wget'):
                vipy.globals.log.info('downloading "%s" to "%s"' % (IMAGENET21K_URL , self._datadir))                
                os.system('wget --no-check-certificate --continue --tries=32 -O %s %s ' % (os.path.join(self._datadir, filetail(IMAGENET21K_URL)), IMAGENET21K_URL))  # server fails many times, need smart continue
            else:
                vipy.downloader.download(IMAGENET21K_URL, os.path.join(self._datadir, filetail(IMAGENET21K_URL)))  # fallback on dumb downloader
            vipy.downloader.unpack(os.path.join(self._datadir, filetail(IMAGENET21K_URL)), self._datadir)  # fallback on dumb downloader

            for f in vipy.util.findtar(os.path.join(datadir, 'winter21_whole')):
                if not os.path.exists(filefull(f)):
                    vipy.downloader.unpack(f, filefull(f), progress=False)
                    os.remove(f)  # cleanup
                
            vipy.downloader.download(IMAGENET21K_WORDNET_ID, os.path.join(self._datadir, 'imagenet21k_wordnet_ids.txt'))
            vipy.downloader.download(IMAGENET21K_WORDNET_LEMMAS, os.path.join(self._datadir, 'imagenet21k_wordnet_lemmas.txt'))

            open(os.path.join(self._datadir, '.complete'), 'a').close()
        
        cachefile = os.path.join(self._datadir, '.imlist.txt')
        if recache and os.path.exists(cachefile):
            os.remove(cachefile)
            
        # Class names: https://github.com/google-research/big_transfer/issues/7
        self._synset_to_categorylist = {x:[y.rstrip().lstrip() for y in lemma.split(',')] for (x,lemma) in zip(vipy.util.readtxt(os.path.join(self._datadir, 'imagenet21k_wordnet_ids.txt')), vipy.util.readtxt(os.path.join(self._datadir, 'imagenet21k_wordnet_lemmas.txt')))}

        f_category = lambda c, synset_to_categorylist=self._synset_to_categorylist, aslemma=aslemma: synset_to_categorylist[c] if aslemma else c
        imglist = vipy.util.findimages(os.path.join(datadir, 'winter21_whole')) if not os.path.exists(cachefile) else vipy.util.readlist(cachefile)
        
        imlist = ([vipy.image.TaggedImage(filename=f, attributes={'wordnet_id':f.rsplit('/',2)[-2]}, tags=f_category(f.rsplit('/',2)[-2])) for f in imglist] if os.name == 'posix' else
                  [vipy.image.TaggedImage(filename=f, attributes={'wordnet_id':os.path.basename(os.path.dirname(f))}, tags=f_category(os.path.basename(os.path.dirname(f)))) for f in imglist])
        super().__init__(imlist, id='imagenet21k')

        if not os.path.exists(cachefile):
            vipy.util.writelist(imlist, cachefile)  # cache me for faster loading instead of walking the directory tree, not relocatable

        
    def synset_to_category(self, s=None):
        return self._synset_to_categorylist if s is None else self._synset_to_categorylist[s]
         
   
class Imagenet2014_DET():
    """Imagenet2014_DET requires login at https://image-net.org from the same IP address as the download, and agreeing to the ImageNet terms: https://image-net.org/download-images.php#term"""    
    def __init__(self, datadir=None, redownload=False, recache=False):    
        datadir = tocache('imagenet2014_det') if datadir is None else datadir
        
        self._datadir = remkdir(datadir)

        if recache:
            if os.path.exists(os.path.join(self._datadir, 'imagenet2014_det_trainset.json')):
                os.remove(os.path.join(self._datadir, 'imagenet2014_det_trainset.json'))
            if os.path.exists(os.path.join(self._datadir, 'imagenet2014_det_valset.json')):
                os.remove(os.path.join(self._datadir, 'imagenet2014_det_valset.json'))
                
        if redownload or not os.path.exists(os.path.join(self._datadir, '.complete')):
            for url in URLS_2014:
                vipy.downloader.download(url, os.path.join(self._datadir, filetail(url)))

            vipy.downloader.download(IMAGENET21K_WORDNET_ID, os.path.join(self._datadir, 'imagenet21k_wordnet_ids.txt'))
            vipy.downloader.download(IMAGENET21K_WORDNET_LEMMAS, os.path.join(self._datadir, 'imagenet21k_wordnet_lemmas.txt'))
                
            for url in URLS_2014:
                vipy.downloader.unpack(os.path.join(self._datadir, filetail(url)), remkdir(os.path.join(self._datadir, filebase(url))))
        
            for f in vipy.util.findtar(os.path.join(self._datadir, 'ILSVRC2014_DET_train')):
                vipy.downloader.unpack(f, filefull(f))

            for f in vipy.util.findtar(os.path.join(self._datadir, 'ILSVRC2014_DET_val')):
                vipy.downloader.unpack(f, filefull(f))

            for f in vipy.util.findtargz(os.path.join(self._datadir, 'bboxes_annotation')):
                vipy.downloader.unpack(f, filefull(f))
                
            open(os.path.join(self._datadir, '.complete'), 'a').close()        

    def synset_to_category(self):
        return {x:[y.rstrip().lstrip() for y in lemma.split(',')] for (x,lemma) in zip(vipy.util.readtxt(os.path.join(self._datadir, 'imagenet21k_wordnet_ids.txt')), vipy.util.readtxt(os.path.join(self._datadir, 'imagenet21k_wordnet_lemmas.txt')))}        
            
    def trainset(self):
        cachefile = os.path.join(self._datadir, 'imagenet2014_det_trainset.json')

        if not os.path.exists(cachefile):
            log.warning('Initial parsing of XML files takes a long time... Consider "pip install lxml ujson" to speed this up')            
            imgfiles = {filebase(f):f for f in vipy.util.findimages(os.path.join(self._datadir, 'ILSVRC2014_DET_train'))}  # slow-ish
            xmlfiles = {filebase(f):f for f in vipy.util.findxml(os.path.join(self._datadir, 'bboxes_annotation'))}  # slow-ish
            xmlfiles |= {filebase(f):f for f in vipy.util.findxml(os.path.join(self._datadir, 'ILSVRC2014_DET_bbox_train'))}  # slow-ish
            xmlfiles |= {filebase(f):f for f in vipy.util.findxml(os.path.join(self._datadir, 'ILSVRC2012_bbox_train_v2'))}  # slow-ish
            d_synset_to_category = self.synset_to_category()
            
            imlist = []        
            for (k,f) in imgfiles.items():
                im = vipy.image.Scene(filename=f)
                if k in xmlfiles and os.path.exists(xmlfiles[k]):
                    root = ET.parse(xmlfiles[k]).getroot()  
                    for obj in root.findall("object"):
                        name = obj.findtext("name")
                        subcategory = obj.findtext("name")                    
                        b = obj.find("bndbox")
                        xmin = int(b.findtext("xmin"))
                        ymin = int(b.findtext("ymin"))
                        xmax = int(b.findtext("xmax"))
                        ymax = int(b.findtext("ymax"))
                        tags = (d_synset_to_category[name] if name in d_synset_to_category else [name]) + (d_synset_to_category[subcategory] if subcategory in d_synset_to_category else [subcategory])
                        im.add_object(vipy.object.Detection(tags=tags, xmin=int(xmin), ymin=int(ymin), xmax=int(xmax), ymax=int(ymax), attributes={'wordnet_id':name, 'subcategory':subcategory}))
                imlist.append(im)                
            vipy.save(imlist, cachefile)  # cached
        else:
            try:
                imlist = vipy.load(cachefile)
            except:
                if os.path.exists(cachefile):
                    os.remove(cachefile)  # force recache on failure
                return self.trainset()
            
        return vipy.dataset.Dataset(imlist, 'imagenet2014_det:trainset')


    def valset(self):
        cachefile = os.path.join(self._datadir, 'imagenet2014_det_valset.json')

        if not os.path.exists(cachefile):
            log.warning('Initial parsing of XML files takes a long time... Consider "pip install lxml ujson" to speed this up')            
            imgfiles = {filebase(f):f for f in vipy.util.findimages(os.path.join(self._datadir, 'ILSVRC2013_DET_val'))}  # slow-ish
            xmlfiles = {filebase(f):f for f in vipy.util.findxml(os.path.join(self._datadir, 'bboxes_annotation'))}  # slow-ish
            xmlfiles |= {filebase(f):f for f in vipy.util.findxml(os.path.join(self._datadir, 'ILSVRC2013_DET_bbox_val'))}  # slow-ish
            xmlfiles |= {filebase(f):f for f in vipy.util.findxml(os.path.join(self._datadir, 'ILSVRC2012_bbox_val_v3'))}  # slow-ish
            d_synset_to_category = self.synset_to_category()
            
            imlist = []        
            for (k,f) in imgfiles.items():
                im = vipy.image.Scene(filename=f)
                if k in xmlfiles and os.path.exists(xmlfiles[k]):
                    root = ET.parse(xmlfiles[k]).getroot()  
                    for obj in root.findall("object"):
                        name = obj.findtext("name")
                        subcategory = obj.findtext("name")                    
                        b = obj.find("bndbox")
                        xmin = int(b.findtext("xmin"))
                        ymin = int(b.findtext("ymin"))
                        xmax = int(b.findtext("xmax"))
                        ymax = int(b.findtext("ymax"))
                        tags = (d_synset_to_category[name] if name in d_synset_to_category else [name]) + (d_synset_to_category[subcategory] if subcategory in d_synset_to_category else [subcategory])
                        im.add_object(vipy.object.Detection(tags=tags, xmin=int(xmin), ymin=int(ymin), xmax=int(xmax), ymax=int(ymax), attributes={'wordnet_id':name, 'subcategory':subcategory}))
                imlist.append(im)                
            vipy.save(imlist, cachefile)  # cached
        else:
            try:
                imlist = vipy.load(cachefile)
            except:
                if os.path.exists(cachefile):
                    os.remove(cachefile)  # force recache on failure
                return self.trainset()
            
        return vipy.dataset.Dataset(imlist, 'imagenet2014_det:valset')


    def testset(self):
        return vipy.dataset.Dataset.from_directory(os.path.join(self._datadir, 'ILSVRC2013_DET_test'), filetype='jpg').id('imagenet2014_det:testset')




import os
from vipy.util import remkdir, tocache
import vipy.downloader
from vipy.dataset import Dataset
from vipy.image import ImageCategory


URL = 'https://data.caltech.edu/records/mzrjq-6wc02/files/caltech-101.zip'
SHA1 = 'd1cc0e3686b03d5e1a7e9b734c6d04f60857d674'


class Caltech101(Dataset):
    """Caltech-101 dataset: https://data.caltech.edu/records/mzrjq-6wc02"""
    def __init__(self, datadir=None, redownload=False):
        """Caltech101, provide a datadir='/path/to/store/caltech101' """
        datadir = tocache('caltech101') if datadir is None else datadir
        
        # Download        
        self._datadir = remkdir(datadir)        
        if redownload or not os.path.exists(os.path.join(self._datadir, '.complete')):
            vipy.downloader.download_and_unpack(URL, self._datadir, sha1=SHA1)            
            vipy.downloader.unpack(os.path.join(self._datadir, 'caltech-101/101_ObjectCategories.tar.gz'), os.path.join(self._datadir, 'caltech-101'))
            
        # Create dataset
        imlist = []
        categorydir = os.path.join(self._datadir, 'caltech-101', '101_ObjectCategories')        
        for (idx_category, category) in enumerate(os.listdir(categorydir)):
            imdir = os.path.join(categorydir, category)
            for imf in os.listdir(imdir):
                imlist.append((category, os.path.join(categorydir, category, imf)))                

        loader = lambda x, categorydir=categorydir: ImageCategory(filename=x[1], category=x[0])                
        super().__init__(imlist, id='caltech-101', loader=loader)

        # Done
        open(os.path.join(self._datadir, '.complete'), 'a').close()
        



import os
import vipy.util
from vipy.util import filetail, remkdir, readjson, groupbyasdict, filefull, readlist, readcsv, filebase
import vipy.downloader
from vipy.video import VideoCategory, Video, Scene
import numpy as np
from vipy.object import Track, BoundingBox, Detection
from vipy.activity import Activity
from vipy.dataset import Dataset


# https://research.google.com/ava/download.html#ava_actions_download
URL = 'https://research.google.com/ava/download/ava_v2.2.zip'


class AVA(object):
    def __init__(self, datadir):
        """AVA, provide a datadir='/path/to/store/ava' """
        self.datadir = remkdir(datadir)
        if not self._isdownloaded():
            self.download()

            
    def __repr__(self):
        return str('<vipy.data.ava_v2.2: "%s">' % self.datadir)

    def download(self):
        zipfile = os.path.join(self.datadir, filetail(URL))
        vipy.downloader.download(URL, zipfile, verbose=False)
        vipy.downloader.unpack(zipfile, self.datadir, verbose=False)
        return self

    def _isdownloaded(self):
        return os.path.exists(os.path.join(self.datadir, 'ava_train_v2.2.csv'))
    
    def _dataset(self, csvfile, downloaded=False, verbose=False):
        # AVA csv format: video_id, middle_frame_timestamp, scaled_person_box (xmin, ymin, xmax, ymax), action_id, person_id

        # video_id: YouTube identifier
        # middle_frame_timestamp: in seconds from the start of the YouTube.
        # person_box: top-left (x1, y1) and bottom-right (x2,y2) normalized with respect to frame size, where (0.0, 0.0) corresponds to the top left, and (1.0, 1.0) corresponds to bottom right.
        # action_id: identifier of an action class, see ava_action_list_v2.2.pbtxt
        # person_id: a unique integer allowing this box to be linked to other boxes depicting the same person in adjacent frames of this video.
        
        assert self._isdownloaded(), "Dataset not downloaded.  download() first or manually download '%s' into '%s'" % (URL, self.datadir)
        csv = readcsv(csvfile)
        d_videoid_to_rows = groupbyasdict(csv, lambda x: x[0])

        vidlist = []
        d_category_to_index = self.categories()
        d_index_to_category = {v:k for (k,v) in d_category_to_index.items()}

        videos = [vipy.video.Scene(url='https://www.youtube.com/watch?v=%s' % video_id, framerate=None, 
                                   filename=os.path.join(self.datadir, video_id)) for (k_video, (video_id, rowlist)) in enumerate(d_videoid_to_rows.items())]                    
        for (k_video, (video_id, rowlist)) in enumerate(d_videoid_to_rows.items()):
            v = videos[k_video]

            if verbose:
                print('[vipy.data.ava][%d/%d]: Parsing "%s" with %d activities' % (k_video, len(d_videoid_to_rows), v.url(), len(rowlist)))            
            dummy_framerate = 30  # placeholder, don't know the true framerate until the video is downloaded

            # Tracks are "actor_id" across the video
            tracks = groupbyasdict(rowlist, lambda x: x[7])
            d_tracknum_to_track = {}

            for (k,(tracknum, tracklist)) in enumerate(tracks.items()):
                (keyframes, boxes) = zip(*[(((float(x[1]))*dummy_framerate), Detection(xmin=float(x[2]), ymin=float(x[3]), xmax=float(x[4]), ymax=float(x[5]), normalized_coordinates=True)) for x in tracklist])
                t = Track(keyframes=keyframes, boxes=boxes, category=tracknum, framerate=dummy_framerate, id=k)
                d_tracknum_to_track[tracknum] = t
                v.add_object(t, rangecheck=False)
                
            # Every row is a separate three second long activity centered at startsec involving one actor
            for (k,(video_id, startsec, xmin, ymin, xmax, ymax, activity_id, actor_id)) in enumerate(rowlist):
                t = d_tracknum_to_track[actor_id]
                try:
                    a = Activity(startframe=max(0, int(np.round(((float(startsec)-1.5)*dummy_framerate)))), endframe=int(np.round(((float(startsec)+1.5)*dummy_framerate))),
                                 category=d_index_to_category[int(activity_id)],
                                 tracks={t.id():t}, framerate=dummy_framerate, id=k)
                    v.add_object(a, rangecheck=False)

                except KeyboardInterrupt:
                    raise
                
                except Exception as e:
                    print('[vipy.data.ava]: actor_id=%s, activity_id=%s, video_id=%s - SKIPPING with error "%s"' % (actor_id, activity_id, video_id, str(e)))                   
                    raise

            start = float(max(0, (min([float(x[1]) for x in rowlist])-1.5)))
            end = float(max([float(x[1]) for x in rowlist])+1.5)
            v = v.clip(start, end)                
            vidlist.append(v)
        return vidlist

    def categories(self):
        rowlist = readlist(os.path.join(self.datadir, 'ava_action_list_v2.2.pbtxt'))
        rowlist = [r.strip() for r in rowlist]  # remove whitespace
        
        d_category_to_index = {}
        for (k,r) in enumerate(rowlist):
            if 'name' in r:

                category = str(r.replace('name: ', '').replace('"',''))
                index = int(rowlist[k+1].replace('label_id: ','').replace('"',''))
                d_category_to_index[category] = index
                
        return d_category_to_index

    def trainset(self):
        return Dataset(self._dataset(os.path.join(self.datadir, 'ava_train_v2.2.csv')), id='ava:train')

    def valset(self):
        return Dataset(self._dataset(os.path.join(self.datadir, 'ava_val_v2.2.csv')), id='ava:val')

    



MAJOR = 1
MINOR = 16
RELEASE = 2


class Version():
    """vipy.version.Version class

    This provides a simple class for comparing version strings similar to packaging.

    >>> ver = vipy.version.Version.from_string('1.2.3')
    >>> ver = vipy.version.Version(major=1, minor=2, release=3)    
    >>> ver >= '1.2.2'
    >>> ver == '1.2.3'
    
    """
    
    def __init__(self, major, minor, release=0):
        self.major = int(major)
        self.minor = int(minor)
        self.release = int(release)

    def __repr__(self):
        return '<vipy.version.Version: %s>' % self.version()

    def __str__(self):
        return self.version()

    def __eq__(self, other):        
        return str(self) == str(Version.from_string(other))

    def __gt__(self, other):
        return self.int()>Version.from_string(other).int()
    
    def __ge__(self, other):
        return self.int()>=Version.from_string(other).int()        

    def __lt__(self, other):
        return self.int()<Version.from_string(other).int()                
    
    def __le__(self, other):
        return self.int()<=Version.from_string(other).int()                        
        
    @classmethod
    def is_valid(cls, versionstr):
        try:
            Version.from_string(versionstr)
            return True
        except:
            return False
        
    @classmethod
    def from_string(cls, version):
        """version string can be 'PACKAGE-X.Y.Z' or 'X.Y.Z' or 'X.Y'"""
        assert isinstance(version, str), "invalid version string"
        assert version.count('.') in [1,2], "invalid version string"
        assert version.rsplit('-',1)[-1].count('.') in [1,2], "invalid version string"                
        
        return cls(*version.rsplit('-',1)[-1].split('.'))

    def int(self):
        return self.major*10000 + self.minor*100 + self.release
    
    def version(self):
        return '%d.%d.%d' % (self.major, self.minor, self.release)        

    def at_least_version(self, v):
        """Is versionstring='X.Y.Z' at least the current version?"""        
        return self.at_least_major_version(v) and self.at_least_minor_version(v) and self.at_least_release_version(v)

    def at_least_major_version(self, v):
        return self.major >= Version.from_string(v).major
    
    def same_major_version(self, v):
        return self.major == Version.from_string(v).major
    
    def at_least_minor_version(self, v):
        return self.minor  >= Version.from_string(v).minor
    
    def at_least_release_version(self, v):
        return self.release >= Version.from_string(v).release

    
VERSION = Version(MAJOR, MINOR, RELEASE)




import os
from vipy.util import readcsv, remkdir
from vipy.image import ImageDetection
import numpy as np


class FaceScrub(object):
    def __init__(self, datadir):
        self._datadir = datadir
        self._dataset = []  # parsed and validated ImageDetections

    def __repr__(self):
        return str('<vipy.data.facescrub: %s>' % self._datadir)

    def __len__(self):
        return len(self._dataset)

    def parse(self):
        """ Return a list of ImageDetections for all URLs in facescrub """
        imset = []
        imdir = remkdir(os.path.join(self._datadir, 'images'))
        csv_actors = readcsv(os.path.join(self._datadir, 'facescrub_actors.txt'), separator='\t')
        for (subjectname, imageid, faceid, url, bbox, sha256) in csv_actors[1:]:
            categoryname = subjectname.replace(' ', '_')
            (xmin,ymin,xmax,ymax) = bbox.split(',')
            imset.append(ImageDetection(url=url, filename=os.path.join(imdir, '%s_%s.jpg' % (categoryname, imageid)), category=categoryname, xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax, attributes={'GENDER':'male'}))

        csv_actresses = readcsv(os.path.join(self._datadir, 'facescrub_actresses.txt'), separator='\t')
        for (subjectname, imageid, faceid, url, bbox, sha256) in csv_actresses[1:]:
            categoryname = subjectname.replace(' ', '_')
            (xmin,ymin,xmax,ymax) = bbox.split(',')
            imset.append(ImageDetection(url=url, filename=os.path.join(imdir, '%s_%s.jpg' % (categoryname, imageid)), category=categoryname, xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax, attributes={'GENDER':'female'}))

        return imset

    def download(self):
        """ Download every URL in dataset and store in provided filename """
        return [d.download(ignoreErrors=True) for d in self.parse()]

    def validate(self):
        """Validate downloaded dataset and store cached list of valid bounding boxes and loadable images accessible with dataset()"""
        P = self.parse()
        D = []
        for (k,p) in enumerate(P):
            if k % 1000 == 0:
                print('[vipy.data.facescrub][%d/%d]: validating dataset... (successful download?, good bounding box?, loadable image?)' % (k, len(P)))
            if not p.invalid() and p.load(ignoreErrors=True, fetch=False) is not None:
                D.append(p.flush())
        self._dataset = D  # cache
        return self

    def dataset(self):
        return self._dataset if len(self) > 0 else self.validate().dataset()

    def stats(self):
        print('[vipy.data.facescrub]: %f percent downloaded' % (float(len(self.parse())) / float(len(self.dataset()))))

    def subjects(self):
        return list(set([im.category() for im in self.dataset()]))

    def split(self, valsize=128):
        D = self.dataset()
        subjects = list(set([im.category() for im in D]))
        (trainset, valset) = ([], [])
        for s in subjects:
            S = [d for d in D if d.category() == s]
            if len(S) > 3:
                valset = valset + S[0:2]   # two examples per subject
                trainset = trainset + S[2:]  # rest
            else:
                trainset = trainset + S
        valset = np.random.choice(valset, valsize)
        return (trainset, valset)
